/* wptRegistry.cpp - W32 Registry access
 *	Copyright (C) 2000-2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>

#include "wptErrors.h"
#include "wptW32API.h"
#include "wptGPG.h"
#include "wptRegistry.h"
#include "wptKeyserver.h"
#include "wptTypes.h"
#include "wptNLS.h"
#include "wptVersion.h"

#define rc_ok(rc) ((rc) == ERROR_SUCCESS)

#define WINPT_REG "Software\\WinPT"

/* GPG file association context. */
struct gpg_filetype {    
    const char *descr;
    const char *ext;
    int nicon;
};

struct reg_hotkey_s reg_hotkeys[] = {
    {"ClipEncrypt", "", 0},
    {"ClipDecrypt", "", 0},
    {"ClipSign",    "", 0},
    {"ClipSignEnc", "", 0},
    {"CwsEncrypt",  "", 0},
    {"CwsDecrypt",  "", 0},
    {"CwsSign",     "", 0},
    {"CwsSignEnc",  "", 0},
    {NULL, "", 0}
};

/* Global WinPT registry prefereneces. */
winpt_reg_prefs_s reg_prefs;


/* Return != 0 if GPG4win is installed. */
int
is_gpg4win_installed (void)
{
    char *p;

    p = get_reg_entry_gpg4win (NULL);
    if (!p)
	return 0;
    if (dir_exist_check (p)) {
	free_if_alloc (p);
	return 0;
    }
    free_if_alloc (p);
    return -1;
}


/* Return != 0 if GPGee is installed. */
int
is_gpgee_installed (void)
{
    HKEY hk;
    LONG ec;

    ec = RegOpenKeyEx (HKEY_CURRENT_USER, "Software\\GPGee", 0, KEY_READ, &hk);
    if (ec == ERROR_SUCCESS) {
	RegCloseKey (hk);
	return -1;
    }

    return 0;
}


/* Free all members in the registry preference struct. */
void
free_reg_prefs (void)
{
    free_if_alloc (reg_prefs.backup.path);
    free_if_alloc (reg_prefs.kserv_conf);
    free_if_alloc (reg_prefs.homedir);
    memset (&reg_prefs, 0, sizeof reg_prefs);
}


/* Register the given WinPT filetype. */
static int
regist_single_filetype (gpg_filetype *gfile)
{
    char icon[256];
    char prog[256];
    
    memset (&icon, 0, sizeof (icon));
    GetModuleFileName (glob_hinst, prog, sizeof (prog)-1);
    _snprintf (icon, sizeof (icon) -1, "%s,%d", prog, gfile->nicon);
    return create_file_type (prog, gfile->ext, gfile->descr, icon);
}


/* Install the GPG related into the W32 resgistry, if the entry already
   exists the function returns immediately. */
int
regist_inst_gnupg (int create_mokey)
{
    int rc;
    HKEY reg;

    rc = RegOpenKeyEx( HKEY_CURRENT_USER, "Software\\GNU\\GnuPG", 0, KEY_READ, &reg );	
    if( rc_ok( rc ) ) {
	RegCloseKey( reg );
	return 0;	
    }
    rc = RegCreateKey( HKEY_CURRENT_USER, "Software\\GNU\\GnuPG", &reg );
    if( !rc_ok( rc ) )		
	return WPTERR_REGISTRY;	
    RegCloseKey( reg );
    if( create_mokey ) {
	rc = RegOpenKeyEx( HKEY_CURRENT_USER,  "Control Panel\\MingW32\\NLS", 0, KEY_READ, &reg );
	if( rc_ok( rc ) ) {
	    RegCloseKey( reg );
	    return 0;	
	}
	rc = RegCreateKey( HKEY_CURRENT_USER, "Control Panel\\MingW32\\NLS", &reg );
	if( !rc_ok( rc ) )
	    return WPTERR_REGISTRY;	
	RegCloseKey( reg );
    }

    return 0;
}


/* Install WinPT into the W32 registry, if the entry already 
   exists the function returns immediately. @with_ext can be
   used to register some file types (if 1). @created contains
   1 if the registry key was created.
   Return value: 0 on success. */
int
regist_inst_winpt (int with_ext, int *created)
{    
    HKEY reg;
    char *p = NULL;
    char modpath[MAX_PATH+1];
    int rc, i, id, n;

    gpg_filetype gpg_filetypes[] = {
	{_("GPG Detached Signature"), ".sig", 1},
	{_("GPG Encrypted Data"),     ".gpg", 2},
	{_("GPG Armored Data"),       ".asc", 2},
	{0}
    };

    if (created)
	*created = 0;
    if (is_gpgee_installed ())
	with_ext = 0;

    rc = RegOpenKeyEx (HKEY_CURRENT_USER, WINPT_REG, 0, KEY_READ, &reg);
    if (rc_ok (rc)) {
	RegCloseKey (reg);
	rc = RegOpenKeyEx (HKEY_CURRENT_USER, WINPT_REG"\\Keyserver", 0, KEY_READ, &reg);
	if (!rc_ok (rc)) {
	    RegCreateKey (HKEY_CURRENT_USER, WINPT_REG"\\Keyserver", &reg);
	    RegCloseKey (reg);
	}
	p = get_reg_entry_keyserver ("Default");
	if (!p) {
	    char buf[16];
	    sprintf (buf, "%d", HKP_PORT);
	    set_reg_entry_keyserver ("Default_Port", buf);
	    set_reg_entry_keyserver ("Default", DEF_HKP_KEYSERVER);
	}
	free_if_alloc (p);
	return 0;
    }
    rc = RegCreateKey (HKEY_CURRENT_USER, WINPT_REG, &reg);
    if (!rc_ok (rc))
	return WPTERR_REGISTRY;
    if (created)
	*created = 1;
    RegCloseKey (reg);
    if (with_ext) {
	id = msg_box (NULL, _("WinPT can register some GPG file types for you so they can "
	                      "be processed with a double click in the explorer.\n"
			      "Do you want to continue?"), _("WinPT"), MB_YESNO|MB_INFO);
	if (id == IDYES) {
	    for (i = 0; gpg_filetypes[i].ext; i++) {
		rc = RegOpenKeyEx (HKEY_CLASSES_ROOT, gpg_filetypes[i].ext, 0, KEY_READ, &reg);
		if (rc_ok (rc)) {
		RegCloseKey (reg);
		    id = log_box (_("WinPT WARNING"), MB_YESNO|MB_INFO, 
				  _("It seems there was already a '%s' file type registered by another application.\n"
				    "Do you want to overwrite it?"), gpg_filetypes[i].ext);
		    if (id == IDNO)
			continue;
		}
		regist_single_filetype (&gpg_filetypes[i]);
	    }
	}
    }
    if ((n=GetModuleFileName (NULL, modpath, MAX_PATH-1)) > 0) {
	while (n-- > 0) {
	    if (modpath[n] == '\\') {
		modpath[n] = 0;
		break;
	    }
	}
	set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, "Install Directory", modpath);
    }
    return 0;
}


/* Create a new filetype in the W32 registry.
   We should really care of errors! Otherwise we can damage the registry! */
int
create_file_type (const char *exefile, const char *ext, 
		  const char *extname, char *iconfile)
{
    int rc;
    HKEY reg = NULL;
    char deficon[256], defexec[256], p_exefile[256];
    
    
    rc = RegCreateKey( HKEY_CLASSES_ROOT, ext, &reg );
    if( rc_ok( rc ) ) 
	rc = RegSetValueEx( reg, NULL, 0, REG_SZ, (byte *)extname, strlen( extname ) );
    if( rc_ok( rc ) )
	rc = RegCloseKey( reg );
    if( rc_ok( rc ) )    
	rc = RegCreateKey( HKEY_CLASSES_ROOT, extname, &reg );
    if( rc_ok( rc ) )
	rc = RegSetValueEx( reg, NULL, 0, REG_SZ, (byte *) extname, strlen( extname ) );
    if( rc_ok( rc ) )
	rc = RegCloseKey( reg );
    if( !rc_ok( rc ) ) {
	rc = WPTERR_REGISTRY;
	goto leave;
    }
    
    memset( &deficon, 0, sizeof deficon );
    _snprintf( deficon, sizeof deficon - 1, "%s\\DefaultIcon", extname );
    memset( &defexec, 0, sizeof defexec );
    _snprintf( defexec, sizeof defexec - 1, "%s\\shell\\open\\command", extname );	
    memset( &p_exefile, 0, sizeof p_exefile );
    _snprintf( p_exefile, sizeof p_exefile - 1, "%s %%1", exefile );
    
    rc = RegCreateKey( HKEY_CLASSES_ROOT, deficon, &reg );
    if( rc_ok( rc ) )
	rc = RegSetValueEx(reg, NULL, 0, REG_SZ, (byte *)iconfile, strlen( iconfile ) );
    if( rc_ok( rc ) )
	rc = RegCloseKey( reg );
    if( rc_ok( rc ) )
	rc = RegCreateKey( HKEY_CLASSES_ROOT, defexec, &reg );
    if( rc_ok( rc ) )
	rc = RegSetValueEx( reg, NULL, 0, REG_SZ, (byte *)p_exefile, strlen( exefile ) );
    if( rc_ok( rc ) )
	rc = RegCloseKey( reg );
    if( !rc_ok( rc ) ) {
	rc = WPTERR_REGISTRY;
	goto leave;	
    }
    
leave:
    if( reg )
	RegCloseKey( reg );
    return rc;
} /* create_file_type */


/* Expand a string with %foo% entries so that %foo% will
   be replaced with its actual value. */
static char *
expand_path (const char *path)
{
    DWORD len;
    char *p;

    len = ExpandEnvironmentStrings (path, NULL, 0);
    if (!len)
	return NULL;
    len += 1;
    p = new char[len+1];
    if (!p)
	return NULL;
    len = ExpandEnvironmentStrings (path, p, len);
    if (!len) {
	free_if_alloc (p);
	return NULL;
    }
    return p; 
}


/* Retrieve a registry entry with the directory given in @dir
   and the key given in @key.
   Return value is the value or NULL otherwise. */
char*
get_reg_entry (HKEY root_key, const char * dir, const char * key)
{
    int rc;
    char text[384] = {0};
    DWORD nbytes, type;
    HKEY reg_key = NULL;
    char * p = NULL;
    
    rc = RegOpenKeyEx (root_key, dir, 0, KEY_QUERY_VALUE, &reg_key);
    if( !rc_ok( rc ) )
	goto leave;
    nbytes = sizeof (text) - 1;
    type = REG_SZ;
    
    rc = RegQueryValueEx (reg_key, key, 0, &type, (BYTE *)&text, &nbytes);
    if (!rc_ok (rc) || !nbytes)
	goto leave;
    
    if (type == REG_EXPAND_SZ && strchr (text, '%'))
	p = expand_path (text);
    else {
	p = new char[nbytes + 1];
	if (!p)
	    BUG (0);
	memcpy (p, text, nbytes);
	p[nbytes] = '\0';
    }
    
leave:
    if (reg_key)
	RegCloseKey (reg_key);
    return p;
}


/* XXX: use REG_EXPAND_SZ to support multi user environments.
        keep this type and do not overwrite it with REG_SZ. */ 

/* Set a registry entry. */
int
set_reg_entry (HKEY root_key, const char *dir, const char *key,
	       const char *value)
{
    int rc = 0;
    HKEY reg_key;

    rc = RegOpenKeyEx( root_key, dir, 0, KEY_WRITE, &reg_key );
    if( !rc_ok( rc ) )
	return WPTERR_REGISTRY;
    rc = RegSetValueEx( reg_key, key, 0, REG_SZ, (BYTE *)value, strlen( value ) );
    if( !rc_ok( rc ) )
	rc = WPTERR_REGISTRY;	
    RegCloseKey( reg_key );
    return rc;
} /* set_reg_entry */


int
set_reg_key( HKEY root_key, const char * dir, const char * key,
	     const char * value )
{
    int rc = 0;
    HKEY reg_key;
    
    rc = RegOpenKeyEx( root_key, dir, 0, KEY_WRITE, &reg_key );
    if( !rc_ok( rc ) )
	return WPTERR_REGISTRY;
    
    rc = RegSetValueEx( reg_key, key, 0, REG_SZ, (BYTE *)value, strlen( value ) );	
    if( !rc_ok( rc ) ) {	
	if ( RegCreateKey( root_key, key, &reg_key ) != ERROR_SUCCESS ) {
	    rc = WPTERR_REGISTRY; 
	    goto leave;
	}
	rc = RegSetValueEx( reg_key, key, 0, REG_SZ, (BYTE *)value, strlen( value ) );	    	
	if ( !rc_ok( rc ) )
	    rc = WPTERR_REGISTRY;
    }
    
leave:
    RegCloseKey( reg_key );
    return rc;
} /* set_reg_key */


int
set_reg_entry_gpg (const char * key, const char * value)
{	
    return set_reg_entry (HKEY_CURRENT_USER, "Software\\GNU\\GnuPG", key, value);
}


int
set_reg_entry_mo (const char * value)
{	
    return set_reg_entry (HKEY_CURRENT_USER, "Control Panel\\Mingw32\\NLS", 
			  "MODir", value);
}


char*
get_reg_entry_gpg (const char *key)
{
    char *p;
    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\GNU\\GnuPG", key);
    if (!p || strlen (p) == 0) {
	free_if_alloc (p);
	return NULL;
    }
    return p;
}


/* Return if possible the GPG4Win installation directory concatenated
   with the string in @path if given. */
char*
get_reg_entry_gpg4win (const char *path)
{
    char *p, *pp;
    p = get_reg_entry (HKEY_LOCAL_MACHINE,
		       "Software\\GNU\\GnuPG", "Install Directory");
    if (!p)
	return NULL;
    if (!path)
	return p;
    pp = new char[strlen (p) + strlen (path) + 4];
    if (!pp)
	BUG (NULL);
    sprintf (pp, "%s\\%s", p, path);
    free_if_alloc (p);
    return pp;
}

    
char*
get_reg_entry_mo (void)
{	
    char *p, *pp;
    const char *lang;

    p = get_reg_entry (HKEY_CURRENT_USER, 
		       "Control Panel\\Mingw32\\NLS", "MODir");
    if (p)
	return p;

    lang = get_gettext_langid ();
    if (!lang)    
	return NULL;	
    pp = new char[strlen ("share\\xxxxx\\locale\\LC_MESSAGES")+8];
    if (!pp)    
	BUG (NULL);
    sprintf (pp, "share\\locale\\%s\\LC_MESSAGES", lang);
    p = get_reg_entry_gpg4win (pp);
    free_if_alloc (pp);
    return p;
}


/* All valid configuration commands. */
static const char *cfg [] = {
    NULL,	
    "CacheTime",
    "WordWrap",
    "FastMode",
    "Viewer",
    "WipeMode",
    "AlwaysTrust",
    "AutoBackup",
    "BackupMode",
    "DisableHotkeys",	
    "NoCompressMultiMedia",	
    "Expert",
    "FMProgressBar",
};


int
set_reg_winpt_single (int id, int val)
{
    char buf[64];
    int rc;

    sprintf (buf, "%d", val);
    rc = set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, cfg[id], buf);
    return rc;
}


int
get_reg_winpt_single (int id)
{
    char * buf = NULL;
    int val = 0;

    buf = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, cfg[id]);
    if (buf && *buf != ' ')
	val = 1;
    else if (!buf)
	val = -1;
    free_if_alloc (buf);
    return val;
}


/* Saves the winpt preferences in the registry. */
int
set_reg_winpt_prefs (winpt_reg_prefs_s * opt)
{
    char buf[128];
    size_t i;
    int rc = 0;

    for (i=1; i < DIM (cfg); i++) {
	switch (i) {
	case CFG_CACHETIME:
	    sprintf (buf, "%d", opt->cache_time);
	    break;
	case CFG_WORDWRAP:
	    sprintf (buf, "%d", opt->word_wrap);
	    break;
	case CFG_WIPEMODE:
	    sprintf (buf, "%d", opt->wipe_mode);
	    break;
	case CFG_FASTMODE:
	    sprintf (buf, "%d", opt->use_tmpfiles);
	    break;
	case CFG_NOZIP_MMEDIA:
	    sprintf (buf, "%d", opt->no_zip_mmedia);
	    break;
	case CFG_VIEWER:
	    sprintf (buf, "%d", opt->use_viewer);
	    break;
	case CFG_ALWAYSTRUST:
	    sprintf (buf, "%d", opt->always_trust);
	    break;
	case CFG_AUTOBACKUP:
	    sprintf (buf, "%d", opt->auto_backup);
	    break;
	case CFG_AUTOBAKMODE:
	    sprintf (buf, "%d", opt->backup.mode);
	    break;
	case CFG_DISHOTKEYS:
	    sprintf (buf, "%d", opt->no_hotkeys);
	    break;
	case CFG_EXPERT:
	    sprintf (buf, "%d", opt->expert);
	    break;

	case CFG_FM_PROGRESS:
	    sprintf (buf, "%d", opt->fm.progress);
	    break;
	}
	rc = set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, cfg[i], buf);
	if (rc)
	    goto leave;
    }

    if (opt->backup.path) {
	rc = set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, "BackupPath",
			    opt->backup.path);
	if (rc)
	    goto leave;
    }

    for (i=0; reg_hotkeys[i].reg_entry; i++) {
	strcpy (buf, " ");
	if (reg_hotkeys[i].enabled)
	    strcpy (buf, reg_hotkeys[i].key);
	rc = set_reg_key (HKEY_CURRENT_USER, WINPT_REG, 
			  reg_hotkeys[i].reg_entry, buf);
	if (rc)
	    break;
    }
    
leave:
    if (rc) {
	msg_box (NULL, _("Could not write to Registry."), _("Preferences"), MB_ERR);
	return rc;
    }
    return 0;
}


int
set_reg_winpt_flag (const char * name, int val)
{
    return set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, name, val? "1" : "0");
} /* set_reg_winpt_flag */


int
get_reg_winpt_flag (const char * name)
{
    char * buf;
    int flag = 0;

    buf = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, name);
    if (buf && buf[0] == '1')
	flag = 1;
    else if (!buf || buf && buf[0] != '0')
	flag = -1;
    free_if_alloc (buf);
    return flag;
} /* get_reg_winpt_flag */


/* Retrieve the winpt preferences from the registry. */
int
get_reg_winpt_prefs (winpt_reg_prefs_s * opt)
{
    char *val = NULL;
    size_t i;

    for (i=1; i < DIM (cfg); i++) {
	val = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, cfg[i]);
	if (!val || *val == ' ') {
	    free_if_alloc (val);
	    continue;
	}
	switch (i) {
	case CFG_CACHETIME:
	    opt->cache_time = atol (val);
	    /* We do NOT support passphrase caching for more than an hour.
	     * Perhaps we should allow it, but for now we silently drop this.
	     */
	    if (opt->cache_time > 3600)
		opt->cache_time = 3600;
	    break;
	case CFG_WORDWRAP:
	    opt->word_wrap = atol (val);
	    break;
	case CFG_FASTMODE:
	    opt->use_tmpfiles = atol (val);
	    break;
	case CFG_NOZIP_MMEDIA:
	    opt->no_zip_mmedia = atol (val);
	    break;
	case CFG_VIEWER:
	    opt->use_viewer = atol (val);
	    break;
	case CFG_WIPEMODE:
	    opt->wipe_mode = atol (val);
	    break;
	case CFG_DISHOTKEYS:
	    opt->no_hotkeys = atol (val);
	    break;
	case CFG_ALWAYSTRUST:
	    opt->always_trust = atol (val);
	    break;
	case CFG_AUTOBACKUP:
	    opt->auto_backup = atol (val);
	    break;
	case CFG_AUTOBAKMODE:
	    opt->backup.mode = atol (val);
	    break;
	case CFG_EXPERT:
	    opt->expert = atol (val);
	    break;

	case CFG_FM_PROGRESS:
	    opt->fm.progress = atol (val);
	    break;
	}
	free_if_alloc (val);
    }    

    val = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, "BackupPath");
    if (val && val[0] != ' ')
	opt->backup.path = m_strdup (val);
    free_if_alloc (val);
    
    for (i=0; reg_hotkeys[i].reg_entry; i++) {
	val = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, reg_hotkeys[i].reg_entry);
	if (val && val[0] != ' ') {
	    reg_hotkeys[i].key[0] = *val;
	    reg_hotkeys[i].enabled = 1;
	}
	else
	    reg_hotkeys[i].enabled = 0;
	free_if_alloc (val);
    }
    return 0;
} /* get_reg_winpt_prefs */


char*
get_reg_entry_keyserver (const char *name)
{
    char * p = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG"\\Keyserver", name);
    if (p && !strcmp (p, "")) {
	free_if_alloc (p);
	return NULL;
    }
    return p;
}


void
get_reg_proxy_prefs (char ** host, int * port, char ** user, char ** pass)
{
    if (host)
	*host = get_reg_entry_keyserver ("Host");
    if (user)
	*user = get_reg_entry_keyserver ("User");
    if (pass)
	*pass = get_reg_entry_keyserver ("Pass");
    if (port) {
	char * p = get_reg_entry_keyserver ("Port");
	if (p) {
	    *port = atol (p);
	    free_if_alloc (p);
	}
    }
} /* get_reg_proxy_prefs */


int
set_reg_entry_keyserver( const char * name, const char * val )
{
    return set_reg_entry( HKEY_CURRENT_USER, WINPT_REG"\\Keyserver", name, val );
} /* set_reg_entry_keyserver */


int
set_reg_proxy_prefs( const char * host, int port, const char * user, const char * pass )
{
    int rc;
    
    rc = set_reg_entry_keyserver( "Host", host? host : "" );
    if( !rc ) {
	char buf[32];
	sprintf( buf, "%d", port );
	rc = set_reg_entry_keyserver( "Port", buf );
    }
    if( !rc )
	rc = set_reg_entry_keyserver( "User", user? user : "" );
    if( !rc )
	rc = set_reg_entry_keyserver( "Pass", pass? pass : "" );
    return rc;
} /* set_reg_proxy_prefs */
